#include "common.h"

struct Appdata
{
    float4 Position	    : POSITION;
    float3 Normal	    : NORMAL;
    float2 Uv	        : TEXCOORD0;
    float2 UvStuds	    : TEXCOORD1;

#ifdef PIN_SKINNED
    float4 ColorDiffuse  : COLOR0;
    int4   ColorSpecular : COLOR1;
#else
    float4 ColorDiffuse  : COLOR0;
    float4 ColorSpecular : COLOR1;
#endif

#ifdef PIN_SURFACE
    float4 SurfaceCoord  : TEXCOORD2;
    float3 Tangent       : TEXCOORD3;
#endif
};

struct VertexOutput
{
    float4 HPosition    : POSITION;
    float2 Uv           : TEXCOORD0;
	float2 UvStuds		: TEXCOORD1;

#ifdef PIN_FFP
    float4 Diffuse      : COLOR0;
    float4 Specular     : COLOR1;
#else
    float4 Color             : COLOR0;

    float4 LightPosition_Fog : TEXCOORD2;

    #if defined(PIN_HQ) || defined(PIN_REFLECTION)
        float4 View_Depth        : TEXCOORD3;
        float4 Normal_SpecPower  : TEXCOORD4;
    #endif
  
    #ifdef PIN_SURFACE
        float4 SurfaceCoord  : TEXCOORD5;
        float4 Tangent_Fade  : TEXCOORD6;
    #else
        float4 Diffuse_Specular  : COLOR1;
    #endif
#endif
};

uniform float3 CameraPosition;
uniform float4x4 ViewProjection;

uniform float3 Lamp0Dir;
uniform float3 Lamp1Dir;
uniform float3 Lamp0Color;
uniform float3 Lamp1Color;
uniform float3 AmbientColor;

uniform float3 FogColor;
uniform float4 FogParams;

uniform float4 LightConfig0;
uniform float4 LightConfig1;
uniform float4 LightBorder;
uniform float2 FadeDistance;

uniform float Reflectance;

#ifdef PIN_SKINNED
uniform float4 WorldMatrixArray[72 * 3];
#endif

#ifdef PIN_DEBUG
uniform float4 DebugColor;
#endif

VertexOutput DefaultVS(Appdata IN)
{
    VertexOutput OUT = (VertexOutput)0;

    // Transform position and normal to world space
#ifdef PIN_SKINNED
    int boneIndex = IN.ColorSpecular.r;

    float4 worldRow0 = WorldMatrixArray[boneIndex * 3 + 0];
    float4 worldRow1 = WorldMatrixArray[boneIndex * 3 + 1];
    float4 worldRow2 = WorldMatrixArray[boneIndex * 3 + 2];
		
	float3 posWorld = float3(dot(worldRow0, IN.Position), dot(worldRow1, IN.Position), dot(worldRow2, IN.Position));
    float3 normalWorld = float3(dot(worldRow0.xyz, IN.Normal), dot(worldRow1.xyz, IN.Normal), dot(worldRow2.xyz, IN.Normal));
#else
	float3 posWorld = IN.Position.xyz;
    float3 normalWorld = IN.Normal;
#endif

    // Decode diffuse/specular parameters; encoding depends on the skinned flag due to vertex declaration differences
#if defined(PIN_DEBUG)
    float4 colorDiffuse = DebugColor;
#else
    float4 colorDiffuse = IN.ColorDiffuse;
#endif

#ifdef PIN_SKINNED
    float specularIntensity = IN.ColorSpecular.g / 255.f;
    float specularPower = IN.ColorSpecular.a;
#else
    float specularIntensity = IN.ColorSpecular.g;
    float specularPower = IN.ColorSpecular.a * 255.f;
#endif

#ifdef PIN_FFP
    // We use a fixed specular power so that vertex processing matches the fixed function vertex processing for static parts
    specularPower = 50.f;

    #ifdef GLSL
        // Due to some issues with FFP implementation of OpenGL & secondary specular, we disable specular stream everywhere
        specularIntensity = 0.9f;
    #endif
#endif

    float ndotl = dot(normalWorld, -Lamp0Dir);

#ifdef PIN_HQ
    // We'll calculate specular in pixel shader
    float2 lt = float2(saturate(ndotl), (ndotl > 0));
#else
    // Using lit here improves performance on software vertex shader implementations
    float2 lt = lit(ndotl, dot(normalize(-Lamp0Dir + normalize(CameraPosition - posWorld.xyz)), normalWorld), specularPower).yz;
#endif

	OUT.HPosition = mul(ViewProjection, float4(posWorld, 1));
	OUT.Uv = IN.Uv;
	OUT.UvStuds = IN.UvStuds;

#ifdef PIN_FFP
    float3 diffuse = AmbientColor + lt.x * Lamp0Color + max(-ndotl, 0) * Lamp1Color;
    float3 specular = lt.y * Lamp0Color;

    OUT.Diffuse = float4(diffuse * colorDiffuse.rgb, colorDiffuse.a);
    OUT.Specular = float4(specular * specularIntensity, 0);
#else
    OUT.Color = colorDiffuse;

    OUT.LightPosition_Fog = float4(lgridPrepareSample(posWorld, normalWorld, LightConfig0, LightConfig1), (FogParams.z - OUT.HPosition.w) * FogParams.w);

    #if defined(PIN_HQ) || defined(PIN_REFLECTION)
        OUT.View_Depth = float4(CameraPosition - posWorld, OUT.HPosition.w);
        OUT.Normal_SpecPower = float4(normalWorld, specularPower);
    #endif

    #ifdef PIN_SURFACE
        OUT.SurfaceCoord = IN.SurfaceCoord;

        #ifdef PIN_SKINNED
            float3 tangent = float3(dot(worldRow0.xyz, IN.Tangent), dot(worldRow1.xyz, IN.Tangent), dot(worldRow2.xyz, IN.Tangent));
        #else
            float3 tangent = IN.Tangent;
        #endif

        OUT.Tangent_Fade = float4(tangent, OUT.HPosition.w * FadeDistance.y);
    #else
        float3 diffuse = lt.x * Lamp0Color + max(-ndotl, 0) * Lamp1Color;

        OUT.Diffuse_Specular = float4(diffuse, lt.y * specularIntensity);
    #endif
#endif

	return OUT;
}

#ifdef PIN_SURFACE
struct Surface
{
    float3 albedo;
    float3 normal;
    float specularIntensity;
    float reflectionIntensity;
};

Surface surfaceShader(VertexOutput IN);
#endif

#ifndef PIN_FFP
sampler2D DiffuseMap: register(s0);
sampler2D StudsMap: register(s1);
sampler3D LightMap: register(s2);
sampler3D NoiseMap: register(s3);
sampler2D NormalMap: register(s4);
samplerCUBE EnvironmentMap: register(s5);

void DefaultPS(VertexOutput IN,
#ifdef PIN_GBUFFER
    out float4 oColor1: COLOR1,
#endif
    out float4 oColor0: COLOR0)
{
    // Compute albedo term
#ifdef PIN_SURFACE
    float4 studs = tex2D(StudsMap, IN.UvStuds);
    Surface surface = surfaceShader(IN);

    float fade = saturate(IN.Tangent_Fade.w);

    surface.albedo = lerp(surface.albedo, IN.Color.rgb, fade);
    surface.normal.xy *= 1 - fade;
    surface.reflectionIntensity *= 1 - fade;

    float4 albedo = float4(lerp(surface.albedo, studs.rgb, studs.a), IN.Color.a);

    float3 bitangent = cross(IN.Normal_SpecPower.xyz, IN.Tangent_Fade.xyz);
    float3 normal = normalize(surface.normal.x * IN.Tangent_Fade.xyz + surface.normal.y * bitangent + surface.normal.z * IN.Normal_SpecPower.xyz);
    float ndotl = dot(normal, -Lamp0Dir);

    float4 diffuseSpecular = float4(saturate(ndotl) * Lamp0Color + max(-ndotl, 0) * Lamp1Color, step(0, ndotl) * surface.specularIntensity);
    float reflectionIntensity = surface.reflectionIntensity;
#else
    float4 diffuseSpecular = IN.Diffuse_Specular;

    #ifdef PIN_REFLECTION
        float reflectionIntensity = 1;
    #endif

    #ifdef PIN_HQ
        float3 normal = normalize(IN.Normal_SpecPower.xyz);
    #elif defined(PIN_REFLECTION)
        float3 normal = IN.Normal_SpecPower.xyz;
    #endif

    #ifdef PIN_PLASTIC
        float4 studs = tex2D(StudsMap, IN.UvStuds);
        float4 albedo = float4(lerp(IN.Color.rgb, studs.rgb, studs.a), IN.Color.a);
    #else
        float4 albedo = tex2D(DiffuseMap, IN.Uv) * IN.Color;
    #endif
#endif

    float4 light = lgridSample(LightMap, IN.LightPosition_Fog.xyz, LightConfig0, LightConfig1, LightBorder);

    // Compute reflection term
#if defined(PIN_SURFACE) || defined(PIN_REFLECTION)
    float3 reflection = reflectionIntensity * texCUBE(EnvironmentMap, reflect(-IN.View_Depth.xyz, normal)).rgb;

    #ifdef PIN_SURFACE
        albedo.rgb += reflection;
    #else
        // Plastic has special rules for how reflection behaves, aargh
        albedo.rgb = lerp(albedo.rgb, reflection.rgb, Reflectance * 0.8);
    #endif
#endif

    // Compute diffuse term
    float3 diffuse = (AmbientColor + diffuseSpecular.rgb * light.a + light.rgb) * albedo.rgb;

    // Compute specular term
#ifdef PIN_HQ
    float3 specular = Lamp0Color * (light.a * diffuseSpecular.a * (float)(half)pow(saturate(dot(normal, normalize(-Lamp0Dir + normalize(IN.View_Depth.xyz)))), IN.Normal_SpecPower.w));
#else
    float3 specular = Lamp0Color * (diffuseSpecular.a * light.a);
#endif

    // Combine
    oColor0.rgb = diffuse.rgb + specular.rgb;
    oColor0.a = albedo.a;

    float fogAlpha = saturate(IN.LightPosition_Fog.w);

#ifdef GLSL
    // Manually apply fog in GLSL path
    oColor0.rgb = lerp(FogColor, oColor0.rgb, fogAlpha);
#endif

#ifdef PIN_GBUFFER
    oColor1 = gbufferPack(IN.View_Depth.w, diffuse.rgb, specular.rgb, fogAlpha);
#endif
}
#endif